#if	defined(__KERNEL__) && defined(_VX_INFO_DEF_)

#include <linux/spinlock.h>
#include <linux/jiffies.h>
#include <asm/param.h>

/* context sub struct */

struct _vx_sched {
	spinlock_t tokens_lock; /* lock for this structure */

	int tokens;		/* number of CPU tokens in this context */
	int tokens_fr;		/* Fill rate: add X tokens... */
	int tokens_div;		/* Divisor:   per Y jiffies   */
	int tokens_max;		/* Limit:     no more than N tokens */
	uint32_t tokens_jfy;    /* add an integral multiple of Y to this */

	uint64_t ticks;		/* token tick events */
};

static inline void vx_info_init_sched(struct _vx_sched *sched)
{
        /* scheduling; hard code starting values as constants */
        sched->tokens_fr 	= 1;
        sched->tokens_div 	= 4;
        sched->tokens     	= HZ * 5;
        sched->tokens_max 	= HZ * 10;
        sched->tokens_jfy 	= jiffies;
        sched->tokens_lock 	= SPIN_LOCK_UNLOCKED;
}

static inline int vx_info_proc_sched(struct _vx_sched *sched, char *buffer)
{
	return sprintf(buffer,
		"TICKS:\t%8lld\n"
		"TOKEN:\t%8d\n"
		"RATE:\t%8d\n"
		"DIV:\t%8d\n"		
		"MAX:\t%8d\n"
		,sched->ticks
		,sched->tokens
		,sched->tokens_fr
		,sched->tokens_div
		,sched->tokens_max
		);
}


#else	/* _VX_INFO_DEF_ */
#ifndef _VX_SCHED_H
#define _VX_SCHED_H

#include <linux/vserver/switch.h>

/*  sched vserver commands */

#define VCMD_set_sched		VC_CMD(SYSTEST, 1, 1)

struct  vcmd_set_sched_v1 {
	int32_t fill_rate;
	int32_t period;
	int32_t fill_level;
	int32_t bucket_size;
};


#ifdef	__KERNEL__

extern int vc_set_sched(uint32_t, void *);


#define VAVAVOOM_RATIO		50

#include <linux/vserver/context.h>


/* scheduling stuff */

int tokens_left(struct task_struct *);
int tokens_recalc(struct task_struct *);

/* update the token allocation for a process */
static inline void tokens_alloc(struct task_struct *tsk)
{
	struct vx_info *vxi = tsk->vx_info;

	if (!vxi)
		return;
	if (!vx_flags(VXF_SCHED_TB, 0))
		return;

	spin_lock(&tsk->vx_info->sched.tokens_lock);
	tokens_recalc(tsk);

	/* else if (eat && tsk->s_info->tokens > 0)
	   {
	   tsk->s_info->tokens--;
	   } */
	spin_unlock(&tsk->vx_info->sched.tokens_lock);
}

/* eat a token, only allocating more if we run out.

   Return:
      1 if a token was eaten,
      0 if the process has no tokens left
     -1 if tokens do not apply
*/
static inline int eat_token(struct task_struct *tsk)
{
	struct vx_info *vxi = tsk->vx_info;
	int eaten = -1;

	if (!vxi)
		goto out;

	spin_lock(&tsk->vx_info->sched.tokens_lock);
	
	vxi->sched.ticks++;
	if (!vx_flags(VXF_SCHED_TB, 0))
		goto out_unlock;

	if (vxi->sched.tokens || tokens_recalc(tsk)) {
		vxi->sched.tokens--;
		eaten = 1;
	} else 
		eaten = 0;

out_unlock:
	spin_unlock(&tsk->vx_info->sched.tokens_lock);
out:
	return eaten;
}

#endif	/* __KERNEL__ */

#endif	/* _VX_SCHED_H */
#endif
